@@ -1,7 +1,8 @@
 #!/usr/bin/env python
 ################ Test Server #################################
-import BaseHTTPServer, cgi
+import BaseHTTPServer, cgi, socket, collections
 from openid2rp import *
+from time import time
 
 # supported providers
 providers = (
@@ -12,7 +13,27 @@ providers = (
     ('Launchpad', 'https://login.launchpad.net/favicon.ico', 'https://login.launchpad.net/')
     )
              
-sessions = []
+
+# Cache discovered information, for later validation
+discovered = {}
+
+# Mapping from OP Endpoint URL to association responses;
+# most recent association is last
+sessions = collections.defaultdict(list)
+# Associations by assoc_handle
+associations = {}
+
+class _Expired(Exception):
+    'Local exception class to indicate expired sessions.'
+    pass
+
+nonces = set()
+def nonce_seen(nonce):
+    if nonce in nonces:
+        return True
+    nonces.add(nonce)
+    return False
+
 class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
 
     def write(self, payload, type):
@@ -38,13 +59,24 @@ class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
                 prov = [p for p in providers if p[0]  == query['provider'][0]]
                 if len(prov) != 1:
                     return self.not_found()
-                prov = prov[0]
-                services, url, op_local = discover(prov[2])
+                prov = prov[0][2]
+                services, url, op_local = discovered[prov] = discover(prov)
+
+                # Get most recent association.  Establish association if it
+                # expired, or if none exist
+                now = time()
                 try:
-                    session = associate(services, url)
-                except ValueError, e:
-                    return self.error(str(e))
-                sessions.append(session)
+                    session, expire = sessions[url][-1]
+                    if now > expire:
+                        raise _Expired
+                except (IndexError, _Expired):
+                    try:
+                        session = associate(services, url)
+                    except ValueError, e:
+                        return self.error(str(e))
+                    sessions[url].append((session, now+int(session['expires_in'])))
+                    associations[session['assoc_handle']] = session
+
                 self.send_response(307) # temporary redirect - do not cache
                 self.send_header("Location", request_authentication
                                  (services, url, session['assoc_handle'],
@@ -65,12 +97,24 @@ class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
                     res = discover(claimed)
                 if res is None:
                     return self.error('Discovery failed')
+                discovered[claimed] = res
                 services, url, op_local = res
+
+                # Get most recent association.  Establish association if it
+                # expired, or if none exist
+                now = time()
                 try:
-                    session = associate(services, url)
-                except ValueError, e:
-                    return self.error(str(e))
-                sessions.append(session)
+                    session, expire = sessions[url][-1]
+                    if now > expire:
+                        raise _Expired
+                except (IndexError, _Expired):
+                    try:
+                        session = associate(services, url)
+                    except ValueError, e:
+                        return self.error(str(e))
+                    sessions[url].append((session, now+int(session['expires_in'])))
+                    associations[session['assoc_handle']] = session
+
                 self.send_response(307)
                 self.send_header("Location", request_authentication
                                  (services, url, session['assoc_handle'],
@@ -79,31 +123,16 @@ class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
                 self.end_headers()
                 return                
             if 'returned' in query:
-                if 'openid.identity' not in query:
+                if 'openid.mode' not in query:
                     return self.rp_discovery()
-                handle = query['openid.assoc_handle'][0]
-                for session in sessions:
-                    if session['assoc_handle'] == handle:
-                        break
-                else:
-                    session = None
-                if not session:
-                    return self.error('Not authenticated (no session)')
                 try:
-                    signed = authenticate(session, querystring)
-                except Exception, e:
-                    self.error("Authentication failed: "+repr(e))
-                    return
-                if 'openid.claimed_id' in query:
-                    if 'claimed_id' not in signed:
-                        return self.error('Incomplete signature')
-                    claimed = query['openid.claimed_id'][0]
-                else:
-                    # OpenID 1, claimed ID not reported - should set cookie
-                    if 'identity' not in signed:
-                        return self.error('Incomplete signature')
-                    claimed = query['openid.identity'][0]
-                payload = "Hello "+claimed+"\n"
+                    signed, claimed_id = verify(query, discovered.get,
+                                                associations.get, nonce_seen)
+                except NotAuthenticated, e:
+                    return self.write('Login failed: %s' % e, 'text/plain')
+                    signed, = query['openid.signed']
+                    signed = signed.split(',')
+                payload = "Hello "+claimed_id+"\n"
                 ax = get_ax(querystring, get_namespaces(querystring), signed)
                 sreg = get_sreg(querystring, signed)
                 email = get_email(querystring)
@@ -169,6 +198,23 @@ class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
     def not_found(self):
         self.send_response(404)
         self.end_headers()
+
+if hasattr(socket, 'IPV6_V6ONLY'):
+    class HTTPServer(BaseHTTPServer.HTTPServer):
+        def __init__(self, addr, handler):
+            if not addr[0]:
+                # use V6 here, set to wildcard below
+                self.address_family = socket.AF_INET6
+            BaseHTTPServer.HTTPServer.__init__(self, addr, handler)
+        def server_bind(self):
+            if self.address_family == socket.AF_INET6:
+                self.socket.setsockopt(socket.IPPROTO_IPV6,
+                                       socket.IPV6_V6ONLY,
+                                       False)
+            BaseHTTPServer.HTTPServer.server_bind(self)
+else:
+    HTTPServer = BaseHTTPServer.HTTPServer
+
         
 # OpenID providers often attempt relying-party discovery
 # This requires the test server to use a globally valid URL
@@ -176,14 +222,19 @@ class Handler(BaseHTTPServer.BaseHTTPRequestHandler):
 # can pass it as command line argument
 def test_server():
     import socket, sys
+    port = 8000
     if len(sys.argv) > 1:
+        import urlparse
         base_url = sys.argv[1]
+        netloc = urlparse.urlparse(base_url).netloc
+        if ':' in netloc:
+            host, port = netloc.split(':')
+            port = int(port)
     else:
         base_url = "http://" + socket.getfqdn() + ":8000/"
     print "Listening on", base_url
     Handler.base_url = base_url
-    #BaseHTTPServer.HTTPServer.address_family = socket.AF_INET6
-    httpd = BaseHTTPServer.HTTPServer(('', 8000), Handler)
+    httpd = HTTPServer(('', port), Handler)
     httpd.serve_forever()
 
 if __name__ == '__main__':
